/*
 * System Interface Library for games
 * Copyright (c) 2007-2016 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * include/SIL/shader.h: Header for graphics shader management.
 */

#ifndef SIL_SHADER_H
#define SIL_SHADER_H

EXTERN_C_BEGIN

/*************************************************************************/
/******************************* Constants *******************************/
/*************************************************************************/

/**
 * ShaderType:  Constants identifying types of shaders, used with the
 * shader creation functions.
 */
enum ShaderType {
    SHADER_TYPE_VERTEX = 0,
    SHADER_TYPE_FRAGMENT,
};
typedef enum ShaderType ShaderType;

/**
 * ShaderAttribute:  Constants identifying standard shader attributes used
 * by the default SIL rendering pipeline.
 */
enum ShaderAttribute {
    SHADER_ATTRIBUTE_POSITION = 0,
    SHADER_ATTRIBUTE_TEXCOORD,
    SHADER_ATTRIBUTE_COLOR,
};
typedef enum ShaderAttribute ShaderAttribute;

/*************************************************************************/
/****************** Interface: Shader object management ******************/
/*************************************************************************/

/**
 * shader_supported:  Return whether the system and the current display
 * mode support shader objects.
 *
 * [Return value]
 *     True if shader objects are supported, false if not.
 */
extern PURE_FUNCTION int shader_supported(void);

/**
 * shader_background_compilation_supported:  Return whether the platform
 * supports compiling shaders to binary code on alternate threads (see
 * the shader_compile_to_binary() documentation).
 *
 * Note that (like all other graphics parameters) the value returned by
 * this function may change after a call to graphics_set_display_mode().
 *
 * [Return value]
 *     True if background shader compilation is supported, false if not.
 */
extern PURE_FUNCTION int shader_background_compilation_supported(void);

/**
 * shader_enable_get_binary:  Indicate whether shader_get_binary() should
 * be supported for subsequently created shaders.  If a true value is
 * passed, shader_get_binary() will succeed on subsequently created shaders
 * (if the system supports it in the first place); otherwise, the behavior
 * of shader_get_binary() on such shaders is undefined.  Calling this
 * function has no effect on shaders created prior to the call.
 *
 * On some systems, the program must declare whether to make binary code
 * available for a shader program when the shader program is created.  By
 * not making this code available, the system may be able to save memory or
 * use a more efficient internal representation or storage method.  This
 * function allows the caller to make that declaration to the system.
 *
 * The default is that shader binary code will not be available, as if
 * shader_enable_get_binary(0) had been called at program startup time.
 *
 * [Parameters]
 *     enable: True to enable shader_get_binary() on subsequent shaders,
 *         false for system-default behavior.
 */
extern void shader_enable_get_binary(int enable);

/**
 * shader_max_attributes:  Return the maximum number of vertex attributes
 * that can be used in a single shader.  The maximum number of attributes
 * is always nonzero (positive) if shader objects are supported.
 *
 * [Return value]
 *     Maximum number of vertex attributes in a shader.
 */
extern PURE_FUNCTION int shader_max_attributes(void);

/**
 * shader_set_attribute:  Define an attribute index-to-name binding for
 * vertex shaders compiled with shader_create_from_source() or
 * shader_compile_to_binary(), such that using GRAPHICS_VERTEX_ATTRIB_*(index)
 * in a vertex format will cause the associated data to appear in the named
 * variable when the shader is executed.  The binding from index to name
 * remains active for all subsequent vertex shader compilations until the
 * index is either redefined (by calling this function with the same index
 * and a different name) or cleared (by calling shader_clear_attributes()).
 * It is an error to attempt to bind the same name to multiple attribute
 * indices.
 *
 * On failure, any previous binding of the given index is cleared.
 *
 * Attributes are saved as part of the compiled shader binary, so they do
 * not need to be set when loading a shader with shader_create_from_binary().
 *
 * It is not required to call this function before compiling a shader if
 * the shader source code defines attribute bindings itself (for example,
 * with the GLSL layout qualifier).  Note, however, that SIL only supports
 * 256 vertex attributes, and any attributes declared with an index of 256
 * or greater will not receive any data.
 *
 * [Parameters]
 *     index: Attribute index; must be in the range [0,255].
 *     name: Attribute name, or NULL to clear any existing binding.
 * [Return value]
 *     True on success, false on error.
 */
extern int shader_set_attribute(int index, const char *name);

/**
 * shader_bind_standard_attribute:  Bind one of the standard vertex
 * attributes to a vertex shader attribute index.
 *
 * By calling this function, a program can associate the standard
 * attribute formats (GRAPHICS_VERTEX_FORMAT_POSITION_* and so on) with
 * specific vertex shader attribute indices, so that primitives containing
 * those format codes will have the corresponding data sent to the
 * specified shader attribute.  In particular, this allows the use of the
 * graphics_begin_primitive() / graphics_add_vertex() /
 * graphics_end_and_draw_primitive() set of functions for immediate
 * primitive drawing, since these functions are hardcoded to use the
 * standard attribute formats.
 *
 * For primitives created or drawn with graphics_create_primitive(),
 * graphics_draw_primitive(), and related functions, if the vertex format
 * contains both a standard vertex attribute and a shader vertex attribute
 * whose index is the same one to which the standard attribute is bound,
 * the data for the standard attribute is ignored and the shader vertex
 * attribute takes precedence.  For example, given the following vertex
 * format and attribute binding:
 *     shader_bind_standard_attribute(SHADER_ATTRIBUTE_POSITION, 0);
 *     typedef struct Vertex {
 *         Vector3f position1;
 *         Vector3f position2;
 *     } Vertex;
 *     uint32_t format[] = {
 *         GRAPHICS_VERTEX_FORMAT(ATTRIB_3F(0), offsetof(Vertex,position1)),
 *         GRAPHICS_VERTEX_FORMAT(POSITION_3F, offsetof(Vertex,position2)),
 *         0
 *     };
 * the vertex shader will receive the data from position1, and position2
 * will be ignored.
 *
 * This function always succeeds.  Passing an out-of-range index (negative
 * or greater than the maximum supported attribute index) disables any
 * existing binding and restores the default behavior of not passing data
 * for that standard attribute to vertex shaders.  Passing an invalid
 * attribute constant has no effect.
 *
 * [Parameters]
 *     attribute: Standard attribute to bind (SHADER_ATTRIBUTE_*).
 *     index: Vertex shader attribute index to bind to, or negative to
 *         cancel any existing binding.
 */
extern void shader_bind_standard_attribute(ShaderAttribute attribute,
                                           int index);

/**
 * shader_clear_attributes:  Clear all attributes previously defined with
 * shader_set_attribute() and all standard attribute bindings defined with
 * shader_set_standard_attribute().
 */
extern void shader_clear_attributes(void);

/**
 * shader_create_from_source:  Create a new shader object by compiling it
 * from the given source code.  The format of the source code is
 * platform-dependent; for OpenGL-based platforms, it should be written as
 * for shader generator callbacks (see graphics_set_shader_generator()).
 *
 * If shader_compilation_supported() returns false, this function will
 * always fail.
 *
 * [Parameters]
 *     type: Type of shader to create (SHADER_TYPE_*).
 *     source: String containing shader source code.
 *     length: Length of source in bytes, or -1 if source is a
 *         null-terminated string.
 * [Return value]
 *     Shader ID (nonzero), or zero on error.
 */
extern int shader_create_from_source(ShaderType type, const char *source,
                                     int length);

/**
 * shader_create_from_binary:  Create a new shader object by loading
 * binary data representing the compiled shader code.  Typically, the
 * data will be obtained by calling shader_get_binary() or
 * shader_compile_to_binary(), but it can also be created using offline
 * shader compilers or similar tools; in that case, see the relevant
 * system-dependent source code, typically the sys_shder_get_binary()
 * implementation, for the proper format of the data buffer to pass to
 * this function.
 *
 * [Parameters]
 *     type: Type of shader to create (SHADER_TYPE_*).
 *     data: Buffer containing compiled shader program data.
 *     size: Size of data, in bytes.
 * [Return value]
 *     Shader ID (nonzero), or zero on error.
 */
extern int shader_create_from_binary(ShaderType type, const void *data,
                                     int size);

/**
 * shader_destroy:  Destroy a shader object.  Does nothing if shader == 0.
 *
 * If the given shader object is bound to a shader pipeline (see
 * shader_pipeline_create()), the shader pipeline object remains live,
 * but attempting to draw primitives with that pipeline applied results
 * in undefined behavior, as if no pipeline was applied.
 *
 * [Parameters]
 *     shader: ID of shader to destroy.
 */
extern void shader_destroy(int shader);

/**
 * shader_get_binary:  Return a buffer containing binary data which can
 * subsequently be passed to shader_create_from_binary().
 *
 * Note that there is generally no guarantee that a shader binary returned
 * by this function can be successfully loaded back into a new shader
 * object.  Even on the same machine, a change of display mode might
 * trigger a change in the display processing pipeline (a change from an
 * integrated GPU to a dedicated graphics card, for example).  Unless a
 * particular platform extrinsically guarantees binary compatibility,
 * callers should always be prepared for shader_create_from_binary() to
 * fail even when the data is valid.
 *
 * The returned buffer should be freed with mem_free() when it is no longer
 * needed.
 *
 * [Parameters]
 *     shader: ID of shader for which to retrieve compiled program data.
 *     size_ret: Pointer to variable to receive the size of the data, in bytes.
 * [Return value]
 *     Binary data representing the compiled shader program, or NULL on error.
 */
extern void *shader_get_binary(int shader, int *size_ret);

/**
 * shader_compile_to_binary:  Compile the given shader source code and
 * return a buffer containing binary data which can subsequently be passed
 * to shader_create_from_binary().  The returned buffer should be freed
 * with mem_free() when it is no longer needed.
 *
 * The caveats which apply to shader_get_binary() also apply to this
 * function.
 *
 * If shader_background_compilation_supported() returns true, then this
 * function may be safely called from any thread.  This can be used, for
 * example, to implement on-demand shader compilation without blocking the
 * rendering thread (as would happen with shader_create_from_source()).
 * Although it is safe to change the display mode while a background
 * compilation thread is live, the caller must ensure that the thread is
 * quiescent during the call to graphics_set_display_mode().  Changing the
 * display mode while a shader is actually being compiled results in
 * undefined behavior.
 *
 * If shader_background_compilation_supported() returns false, then (like
 * other graphics functions) calling this function from any thread other
 * than the main thread results in undefined behavior.
 *
 * [Parameters]
 *     type: Type of shader to create (SHADER_TYPE_*).
 *     source: String containing shader source code.
 *     length: Length of source in bytes, or -1 if source is a
 *         null-terminated string.
 *     size_ret: Pointer to variable to receive the size of the data, in bytes.
 * [Return value]
 *     Binary data representing the compiled shader program, or NULL on error.
 */
extern void *shader_compile_to_binary(ShaderType type, const char *source,
                                      int length, int *size_ret);

/**
 * shader_get_uniform_id:  Return a value identifying the given uniform in
 * the given shader, for use in shader_set_uniform_*() calls.
 *
 * On some platforms (notably OpenGL before 4.1 and OpenGL ES before 3.1),
 * it may not be possible to determine whether a uniform exists in a shader
 * before the shader is actually used for rendering.  In such cases, this
 * function will return a valid ID even for uniforms that do not exist, but
 * the values set for such uniforms will be ignored.  Note that each
 * uniform queried or set may take up additional memory even if the uniform
 * does not exist in the shader, since the names and values must be stored
 * until the uniforms can be looked up.
 *
 * [Parameters]
 *     shader: Shader ID.
 *     name: Name of uniform (e.g., variable name used in shader source code).
 * [Return value]
 *     Uniform ID, or zero if the uniform is not found.
 */
extern int shader_get_uniform_id(int shader, const char *name);

/**
 * shader_set_uniform_*:  Set the value of a shader uniform.  The data type
 * must match the type used in the shader, or undefined behavior results.
 *
 * The value of a texture sampler uniform is the index of the texture unit
 * it accesses (as in the unit parameter to texture_apply()), and the data
 * type is int for the purpose of these functions.
 *
 * For any two shaders used in a pipeline, if both shaders declare a
 * uniform of the same name, it is undefined whether they share storage
 * (and thus always have the same value) or have separate storage (and
 * thus independent values).  To be safe, always ensure that same-named
 * uniforms in a vertex/fragment shader pair are set to the same value.
 *
 * [Parameters]
 *     shader: Shader ID.
 *     uniform: Uniform ID, as returned from shader_get_uniform_id().
 *     value: Value to set.
 */
extern void shader_set_uniform_int(int shader, int uniform, int value);
extern void shader_set_uniform_float(int shader, int uniform, float value);
extern void shader_set_uniform_vec2(
    int shader, int uniform, const struct Vector2f *value);
extern void shader_set_uniform_vec3(
    int shader, int uniform, const struct Vector3f *value);
extern void shader_set_uniform_vec4(
    int shader, int uniform, const struct Vector4f *value);
extern void shader_set_uniform_mat4(
    int shader, int uniform, const struct Matrix4f *value);

/*************************************************************************/
/***************** Interface: Shader pipeline management *****************/
/*************************************************************************/

/**
 * shader_pipeline_create:  Create a new shader pipeline containing the
 * given shaders.  A shader pipeline encapsulates a vertex/fragment shader
 * pair used to render primitives, and is equivalent to (for example) an
 * OpenGL 4.1 program pipeline.
 *
 * On systems which do not support independent vertex and fragment shader
 * programs, this creates and links a new shader program containing the
 * given vertex and fragment shaders, so the same interface may be used
 * without regard to system-dependent details.  Note, however, that on
 * such systems, uniform updates can be expensive because each update must
 * be propagated to all shader programs which use the updated shader.
 *
 * [Parameters]
 *     vertex_shader: ID of vertex shader to use in pipeline.
 *     fragment_shader: ID of fragment shader to use in pipeline.
 * [Return value]
 *     Shader pipeline ID (nonzero), or zero on error.
 */
extern int shader_pipeline_create(int vertex_shader, int fragment_shader);

/**
 * shader_pipeline_destroy:  Destroy a shader pipeline.  Does nothing if
 * pipeline == 0.
 *
 * [Parameters]
 *     pipeline: ID of shader pipeline to destroy.
 */
extern void shader_pipeline_destroy(int pipeline);

/**
 * shader_pipeline_apply:  Use the given shader pipeline for subsequent
 * draw operations.
 *
 * It is permitted to pass zero, which removes any currently applied
 * shader pipeline, but attempting to draw primitives without a shader
 * pipeline applied (assuming shader objects have been enabled with
 * graphics_use_shader_objects()) results in undefined behavior.
 *
 * [Parameters]
 *     pipeline: ID of shader pipeline to apply, or zero to remove the
 *         currently applied shader pipeline.
 */
extern void shader_pipeline_apply(int pipeline);

/*************************************************************************/
/***************************** C++ wrappers ******************************/
/*************************************************************************/

#ifdef __cplusplus
EXTERN_C_END

/*
 * The following macro wraps a function taking an input pointer parameter
 * such that the function can also be called (from C++ only) using a
 * non-pointer value of the same type.  For example:
 *    void shader_set_uniform_mat4(const struct Matrix4f *matrix)
 * is wrapped by:
 *    void shader_set_uniform_mat4(const struct Matrix4f &matrix)
 * allowing a call of the format:
 *    shader_set_uniform_mat4(Matrix4f(...));
 *
 * This is intended mainly to work around the lack of C99-style compound
 * literals in C++.
 */

#define WRAP_FUNCTION_2(NAME,TYPE0,TYPE1,TYPE) \
    static inline void NAME(TYPE0 arg0, TYPE1 arg1, const struct TYPE &value) {NAME(arg0, arg1, &value);}

WRAP_FUNCTION_2(shader_set_uniform_vec2, int, int, Vector2f)
WRAP_FUNCTION_2(shader_set_uniform_vec3, int, int, Vector3f)
WRAP_FUNCTION_2(shader_set_uniform_vec4, int, int, Vector4f)
WRAP_FUNCTION_2(shader_set_uniform_mat4, int, int, Matrix4f)

#undef WRAP_FUNCTION_2

EXTERN_C_BEGIN
#endif  // __cplusplus

/*************************************************************************/
/*************************************************************************/

EXTERN_C_END

#endif  // SIL_SHADER_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
